@@ -1,45 +1,21 @@
-# -*- coding: utf-8 -*-
-#
-# Copyright (c) 2011-2018, Cédric Krier
-# Copyright (c) 2013-2014, Nicolas Évrard
-# Copyright (c) 2011-2018, B2CK
-# All rights reserved.
-#
-# Redistribution and use in source and binary forms, with or without
-# modification, are permitted provided that the following conditions are met:
-#     * Redistributions of source code must retain the above copyright
-#       notice, this list of conditions and the following disclaimer.
-#     * Redistributions in binary form must reproduce the above copyright
-#       notice, this list of conditions and the following disclaimer in the
-#       documentation and/or other materials provided with the distribution.
-#     * Neither the name of the <organization> nor the
-#       names of its contributors may be used to endorse or promote products
-#       derived from this software without specific prior written permission.
-#
-# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-# ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
-# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
-# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
-# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
-# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
-# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
-# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-
-
-
+# This file is part of python-sql.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
+import numbers
 import string
 import warnings
-from threading import local, currentThread
 from collections import defaultdict
 from itertools import chain
+from threading import current_thread, local
 
-__version__ = '1.1.0'
+__version__ = '1.4.3'
 __all__ = ['Flavor', 'Table', 'Values', 'Literal', 'Column', 'Join',
     'Asc', 'Desc', 'NullsFirst', 'NullsLast', 'format2numeric']
 
 
+def _escape_identifier(name):
+    return '"%s"' % name.replace('"', '""')
+
+
 def alias(i, letters=string.ascii_lowercase):
     '''
     Generate a unique alias based on integer
@@ -72,13 +48,16 @@ class Flavor(object):
         paramstyle - state the type of parameter marker formatting
         ilike - support ilike extension
         no_as - doesn't support AS keyword for column and table
+        no_boolean - doesn't support boolean type
         null_ordering - support NULL ordering
         function_mapping - dictionary with Function to replace
+        filter_ - support filter on aggregate functions
+        escape_empty - support empty escape
     '''
 
     def __init__(self, limitstyle='limit', max_limit=None, paramstyle='format',
             ilike=False, no_as=False, no_boolean=False, null_ordering=True,
-            function_mapping=None, filter_=False):
+            function_mapping=None, filter_=False, escape_empty=False):
         assert limitstyle in ['fetch', 'limit', 'rownum']
         self.limitstyle = limitstyle
         self.max_limit = max_limit
@@ -89,6 +68,7 @@ class Flavor(object):
         self.null_ordering = null_ordering
         self.function_mapping = function_mapping or {}
         self.filter_ = filter_
+        self.escape_empty = escape_empty
 
     @property
     def param(self):
@@ -100,7 +80,7 @@ class Flavor(object):
     @staticmethod
     def set(flavor):
         '''Set this thread's flavor to flavor.'''
-        currentThread().__sql_flavor__ = flavor
+        current_thread().__sql_flavor__ = flavor
 
     @staticmethod
     def get():
@@ -111,10 +91,10 @@ class Flavor(object):
         sets this thread's flavor.
         '''
         try:
-            return currentThread().__sql_flavor__
+            return current_thread().__sql_flavor__
         except AttributeError:
             flavor = Flavor()
-            currentThread().__sql_flavor__ = flavor
+            current_thread().__sql_flavor__ = flavor
             return flavor
 
 
@@ -159,6 +139,14 @@ class AliasManager(object):
             return ''
         return cls.local.alias[id(from_)]
 
+    @classmethod
+    def contains(cls, from_):
+        if getattr(cls.local, 'alias', None) is None:
+            return False
+        if from_ in cls.local.exclude:
+            return False
+        return id(from_) in cls.local.alias
+
     @classmethod
     def set(cls, from_, alias):
         assert cls.local.alias.get(from_) is None
@@ -250,6 +238,10 @@ class FromItem(object):
     def alias(self):
         return AliasManager.get(self)
 
+    @property
+    def has_alias(self):
+        return AliasManager.contains(self)
+
     def __getattr__(self, name):
         if name.startswith('__'):
             raise AttributeError
@@ -265,6 +257,27 @@ class FromItem(object):
     def join(self, right, type_='INNER', condition=None):
         return Join(self, right, type_=type_, condition=condition)
 
+    def left_join(self, right, condition=None):
+        return self.join(right, type_='LEFT', condition=condition)
+
+    def left_outer_join(self, right, condition=None):
+        return self.join(right, type_='LEFT OUTER', condition=condition)
+
+    def right_join(self, right, condition=None):
+        return self.join(right, type_='RIGHT', condition=condition)
+
+    def right_outer_join(self, right, condition=None):
+        return self.join(right, type_='RIGHT OUTER', condition=condition)
+
+    def full_join(self, right, condition=None):
+        return self.join(right, type_='FULL', condition=condition)
+
+    def full_outer_join(self, right, condition=None):
+        return self.join(right, type_='FULL OUTER', condition=condition)
+
+    def cross_join(self, right, condition=None):
+        return self.join(right, type_='CROSS', condition=condition)
+
     def lateral(self):
         return Lateral(self)
 
@@ -348,7 +361,7 @@ class SelectQuery(WithQuery):
     @limit.setter
     def limit(self, value):
         if value is not None:
-            assert isinstance(value, int)
+            assert isinstance(value, numbers.Integral)
         self._limit = value
 
     @property
@@ -358,7 +371,7 @@ class SelectQuery(WithQuery):
     @offset.setter
     def offset(self, value):
         if value is not None:
-            assert isinstance(value, int)
+            assert isinstance(value, numbers.Integral)
         self._offset = value
 
     @property
@@ -390,11 +403,11 @@ class SelectQuery(WithQuery):
 
 class Select(FromItem, SelectQuery):
     __slots__ = ('_columns', '_where', '_group_by', '_having', '_for_',
-        'from_', '_distinct', '_distinct_on')
+        'from_', '_distinct', '_distinct_on', '_windows')
 
     def __init__(self, columns, from_=None, where=None, group_by=None,
             having=None, for_=None, distinct=False, distinct_on=None,
-            **kwargs):
+            windows=None, **kwargs):
         self._distinct = False
         self._distinct_on = []
         self._columns = None
@@ -402,6 +415,7 @@ class Select(FromItem, SelectQuery):
         self._group_by = None
         self._having = None
         self._for_ = None
+        self._windows = []
         super(Select, self).__init__(**kwargs)
         self.distinct = distinct
         self.distinct_on = distinct_on
@@ -411,6 +425,7 @@ class Select(FromItem, SelectQuery):
         self.group_by = group_by
         self.having = having
         self.for_ = for_
+        self.windows = windows
 
     @property
     def distinct(self):
@@ -487,6 +502,34 @@ class Select(FromItem, SelectQuery):
             assert all(isinstance(f, For) for f in value)
         self._for_ = value
 
+    @property
+    def windows(self):
+        from sql.aggregate import Aggregate
+        from sql.functions import WindowFunction
+        windows = set()
+        if self._windows is not None:
+            for window in self._windows:
+                windows.add(window)
+                yield window
+        for column in self.columns:
+            window_function = None
+            if isinstance(column, (WindowFunction, Aggregate)):
+                window_function = column
+            elif (isinstance(column, As)
+                    and isinstance(column.expression,
+                        (WindowFunction, Aggregate))):
+                window_function = column.expression
+            if (window_function and window_function.window
+                    and window_function.window not in windows):
+                windows.add(window_function.window)
+                yield window_function.window
+
+    @windows.setter
+    def windows(self, value):
+        if value is not None:
+            assert all(isinstance(w, Window) for w in value)
+        self._windows = value
+
     @staticmethod
     def _format_column(column):
         if isinstance(column, As):
@@ -504,23 +547,6 @@ class Select(FromItem, SelectQuery):
             else:
                 return str(column)
 
-    def _window_functions(self):
-        from sql.functions import WindowFunction
-        from sql.aggregate import Aggregate
-        windows = set()
-        for column in self.columns:
-            window_function = None
-            if isinstance(column, (WindowFunction, Aggregate)):
-                window_function = column
-            elif (isinstance(column, As)
-                    and isinstance(column.expression,
-                        (WindowFunction, Aggregate))):
-                window_function = column.expression
-            if (window_function and window_function.window
-                    and window_function.window not in windows):
-                windows.add(window_function.window)
-                yield window_function
-
     def _rownum(self, func):
         aliases = [c.output_name if isinstance(c, As) else None
             for c in self.columns]
@@ -568,6 +594,13 @@ class Select(FromItem, SelectQuery):
                 from_ = ' FROM %s' % self.from_
             else:
                 from_ = ''
+
+            # format window before expressions to set alias
+            window = ', '.join(
+                '"%s" AS (%s)' % (w.alias, w) for w in self.windows)
+            if window:
+                window = ' WINDOW ' + window
+
             if self.distinct:
                 distinct = 'DISTINCT '
                 if self.distinct_on:
@@ -588,11 +621,6 @@ class Select(FromItem, SelectQuery):
             having = ''
             if self.having:
                 having = ' HAVING ' + str(self.having)
-            window = ''
-            windows = [f.window for f in self._window_functions()]
-            if windows:
-                window = ' WINDOW ' + ', '.join(
-                    '"%s" AS (%s)' % (w.alias, w) for w in windows)
             for_ = ''
             if self.for_ is not None:
                 for_ = ' ' + ' '.join(map(str, self.for_))
@@ -607,25 +635,30 @@ class Select(FromItem, SelectQuery):
                 and (self.limit is not None or self.offset is not None)):
             return self._rownum(lambda q: q.params)
         p = []
-        p.extend(self._with_params())
-        for column in chain(self.distinct_on or (), self.columns):
-            if isinstance(column, As):
-                p.extend(column.expression.params)
-            p.extend(column.params)
-        if self.from_ is not None:
-            p.extend(self.from_.params)
-        if self.where:
-            p.extend(self.where.params)
-        if self.group_by:
-            for expression in self.group_by:
-                p.extend(expression.params)
-        if self.order_by:
-            for expression in self.order_by:
-                p.extend(expression.params)
-        if self.having:
-            p.extend(self.having.params)
-        for window_function in self._window_functions():
-            p.extend(window_function.window.params)
+        with AliasManager():
+            # Set alias to window function to skip their params
+            for window in self.windows:
+                window.alias
+
+            p.extend(self._with_params())
+            for column in chain(self.distinct_on or (), self.columns):
+                if isinstance(column, As):
+                    p.extend(column.expression.params)
+                p.extend(column.params)
+            if self.from_ is not None:
+                p.extend(self.from_.params)
+            if self.where:
+                p.extend(self.where.params)
+            if self.group_by:
+                for expression in self.group_by:
+                    p.extend(expression.params)
+            if self.order_by:
+                for expression in self.order_by:
+                    p.extend(expression.params)
+            if self.having:
+                p.extend(self.having.params)
+            for window in self.windows:
+                p.extend(window.params)
         return tuple(p)
 
 
@@ -701,22 +734,22 @@ class Insert(WithQuery):
         columns = ''
         if self.columns:
             assert all(col.table == self.table for col in self.columns)
-            columns = ' (' + ', '.join(map(str, self.columns)) + ')'
-        if isinstance(self.values, Query):
-            values = ' %s' % str(self.values)
-            # TODO manage DEFAULT
-        elif self.values is None:
-            values = ' DEFAULT VALUES'
+            # Get columns without alias
+            columns = ', '.join(c.column_name for c in self.columns)
+            columns = ' (' + columns + ')'
         with AliasManager():
-            table = self.table
-            AliasManager.set(table, str(table)[1:-1])
+            if isinstance(self.values, Query):
+                values = ' %s' % str(self.values)
+                # TODO manage DEFAULT
+            elif self.values is None:
+                values = ' DEFAULT VALUES'
             returning = ''
             if self.returning:
                 returning = ' RETURNING ' + ', '.join(
                     map(self._format, self.returning))
             return (self._with_str()
-                + 'INSERT INTO %s' % self.table + columns
-                + values + returning)
+                + 'INSERT INTO %s AS "%s"' % (self.table, self.table.alias)
+                + columns + values + returning)
 
     @property
     def params(self):
@@ -766,16 +799,12 @@ class Update(Insert):
     def __str__(self):
         assert all(col.table == self.table for col in self.columns)
         # Get columns without alias
-        columns = list(map(str, self.columns))
+        columns = [c.column_name for c in self.columns]
 
         with AliasManager():
             from_ = ''
             if self.from_:
-                table = From([self.table])
                 from_ = ' FROM %s' % str(self.from_)
-            else:
-                table = self.table
-                AliasManager.set(table, str(table)[1:-1])
             values = ', '.join('%s = %s' % (c, self._format(v))
                 for c, v in zip(columns, self.values))
             where = ''
@@ -786,8 +815,8 @@ class Update(Insert):
                 returning = ' RETURNING ' + ', '.join(
                     map(self._format, self.returning))
             return (self._with_str()
-                + 'UPDATE %s SET ' % table + values + from_
-                + where + returning)
+                + 'UPDATE %s AS "%s" SET ' % (self.table, self.table.alias)
+                + values + from_ + where + returning)
 
     @property
     def params(self):
@@ -853,6 +882,17 @@ class Delete(WithQuery):
             assert isinstance(value, list)
         self._returning = value
 
+    @staticmethod
+    def _format(value, param=None):
+        if param is None:
+            param = Flavor.get().param
+        if isinstance(value, Expression):
+            return str(value)
+        elif isinstance(value, Select):
+            return '(%s)' % value
+        else:
+            return param
+
     def __str__(self):
         with AliasManager(exclude=[self.table]):
             only = ' ONLY' if self.only else ''
@@ -861,7 +901,8 @@ class Delete(WithQuery):
                 where = ' WHERE ' + str(self.where)
             returning = ''
             if self.returning:
-                returning = ' RETURNING ' + ', '.join(map(str, self.returning))
+                returning = ' RETURNING ' + ', '.join(
+                    map(self._format, self.returning))
             return (self._with_str()
                 + 'DELETE FROM%s %s' % (only, self.table)
                 + where + returning)
@@ -891,17 +932,21 @@ class CombiningQuery(FromItem, SelectQuery):
     def __str__(self):
         with AliasManager():
             operator = ' %s %s' % (self._operator, 'ALL ' if self.all_ else '')
-            return (operator.join(map(str, self.queries)) + self._order_by_str
+            return (
+                self._with_str()
+                + operator.join(map(str, self.queries)) + self._order_by_str
                 + self._limit_offset_str)
 
     @property
     def params(self):
         p = []
-        for q in self.queries:
-            p.extend(q.params)
-        if self.order_by:
-            for expression in self.order_by:
-                p.extend(expression.params)
+        with AliasManager():
+            p.extend(self._with_params())
+            for q in self.queries:
+                p.extend(q.params)
+            if self.order_by:
+                for expression in self.order_by:
+                    p.extend(expression.params)
         return tuple(p)
 
 
@@ -937,13 +982,8 @@ class Table(FromItem):
         self._database = database
 
     def __str__(self):
-        if self._database:
-            return '"%s"."%s"."%s"' % (
-                self._database, self._schema, self._name)
-        elif self._schema:
-            return '"%s"."%s"' % (self._schema, self._name)
-        else:
-            return '"%s"' % self._name
+        return '.'.join(map(_escape_identifier, filter(None,
+                    (self._database, self._schema, self._name))))
 
     @property
     def params(self):
@@ -1040,6 +1080,10 @@ class Join(FromItem):
     def alias(self):
         raise AttributeError
 
+    @property
+    def has_alias(self):
+        raise AttributeError
+
     def __getattr__(self, name):
         raise AttributeError
 
@@ -1316,17 +1360,18 @@ class Column(Expression):
     def name(self):
         return self._name
 
+    @property
+    def column_name(self):
+        return (
+            self._name if self._name == '*'
+            else _escape_identifier(self._name))
+
     def __str__(self):
-        if self._name == '*':
-            t = '%s'
-        else:
-            t = '"%s"'
         alias = self._from.alias
         if alias:
-            t = '"%s".' + t
-            return t % (alias, self._name)
+            return '%s.%s' % (_escape_identifier(alias), self.column_name)
         else:
-            return t % self._name
+            return self.column_name
 
     @property
     def params(self):
@@ -1342,7 +1387,7 @@ class As(Expression):
         self.output_name = output_name
 
     def __str__(self):
-        return '"%s"' % self.output_name
+        return '%s' % _escape_identifier(self.output_name)
 
     @property
     def params(self):
@@ -1398,10 +1443,11 @@ class Collate(Expression):
 
 
 class Window(object):
-    __slots__ = ('_partition', '_order_by', '_frame', '_start', '_end')
+    __slots__ = (
+        '_partition', '_order_by', '_frame', '_start', '_end', '_exclude')
 
     def __init__(self, partition, order_by=None,
-            frame=None, start=None, end=0):
+            frame=None, start=None, end=0, exclude=None):
         super(Window, self).__init__()
         self._partition = None
         self._order_by = None
@@ -1413,6 +1459,7 @@ class Window(object):
         self.frame = frame
         self.start = start
         self.end = end
+        self.exclude = exclude
 
     @property
     def partition(self):
@@ -1442,7 +1489,7 @@ class Window(object):
     @frame.setter
     def frame(self, value):
         if value:
-            assert value in ['RANGE', 'ROWS']
+            assert value in ['RANGE', 'ROWS', 'GROUPS']
         self._frame = value
 
     @property
@@ -1452,7 +1499,7 @@ class Window(object):
     @start.setter
     def start(self, value):
         if value:
-            assert isinstance(value, int)
+            assert isinstance(value, numbers.Integral)
         self._start = value
 
     @property
@@ -1462,13 +1509,27 @@ class Window(object):
     @end.setter
     def end(self, value):
         if value:
-            assert isinstance(value, int)
+            assert isinstance(value, numbers.Integral)
         self._end = value
 
+    @property
+    def exclude(self):
+        return self._exclude
+
+    @exclude.setter
+    def exclude(self, value):
+        if value:
+            assert value in ['CURRENT ROW', 'GROUP', 'TIES']
+        self._exclude = value
+
     @property
     def alias(self):
         return AliasManager.get(self)
 
+    @property
+    def has_alias(self):
+        return AliasManager.contains(self)
+
     def __str__(self):
         partition = ''
         if self.partition:
@@ -1492,7 +1553,10 @@ class Window(object):
             start = format(self.start, 'PRECEDING')
             end = format(self.end, 'FOLLOWING')
             frame = ' %s BETWEEN %s AND %s' % (self.frame, start, end)
-        return partition + order_by + frame
+        exclude = ''
+        if self.exclude:
+            exclude = ' EXCLUDE %s' % self.exclude
+        return partition + order_by + frame + exclude
 
     @property
     def params(self):
